"""Functions to enable the app to know what the status of the phone is by
parsing screen data."""
from typing import TYPE_CHECKING, List

import networkx as nx
from typeguard import typechecked

# pylint: disable=R0801
if TYPE_CHECKING:
    from appcommander.Screen import Screen
    from appcommander.Script import Script
else:
    Screen = object
    Script = object


def get_expected_screen_nrs(
    G: nx.DiGraph, screen_nr: int, action_nr: int
) -> List[int]:
    """Returns the expected screens per screen per action."""
    expected_screens: List[int] = []
    for edge in G.edges:
        if edge[0] == screen_nr:
            if action_nr in G[edge[0]][edge[1]]["actions"]:
                expected_screens.append(edge[1])
    return expected_screens


@typechecked
def get_expected_screens(
    expected_screennames: List[int], script_graph: nx.DiGraph
) -> List[Screen]:
    """Determines whether the current screen is one of the expected screens."""
    expected_screens: List[Screen] = []
    for nodename in script_graph.nodes:
        if nodename in expected_screennames:
            expected_screens.append(script_graph.nodes[nodename]["Screen"])

    return expected_screens



"""Plots the script flow."""
import os
from pathlib import Path
from typing import Dict, List, Tuple, Union

import matplotlib.pyplot as plt
import networkx as nx
import PIL
from matplotlib.pyplot import imread
from PIL import Image
from typeguard import typechecked


def visualise_script_flow(
    G: nx.DiGraph, package_name: str, app_version: str
) -> None:
    """Visualises the script flow."""
    img_width, img_height = get_existing_image_size(
        G=G,
        package_name=package_name,
        app_version=app_version,
    )
    resize_image(
        "Unknown_screen.png", target_width=img_width, target_height=img_height
    )

    G.nodes[0]["pos"] = [0, 0]
    set_cyclical_node_coords(evaluated_nodes=[], G=G, start_nodename=0, y=0)
    retry_plot_coordinated_graph(
        app_version=app_version,
        G=G,
        package_name=package_name,
    )


# pylint: disable=R0914
@typechecked
def retry_plot_coordinated_graph(
    app_version: str,
    G: Union[nx.Graph, nx.DiGraph],
    package_name: str,
) -> None:
    """Plots the networkx graph with images instead of nodes.

    :param G: The original graph on which the MDSA algorithm is ran.
    """
    pos: Dict[int, List[int]] = {}
    for nodename in G.nodes:
        pos[nodename] = G.nodes[nodename]["pos"]
        if pos[nodename][1] > 0:
            pos[nodename][1] = pos[nodename][1] - 1
        print(f"{nodename}: {pos[nodename]}")
    # TODO: normalise towards center,
    # TODO: remove skipped horizontal positions.

    fig = plt.figure(figsize=(1, 1))
    ax = plt.subplot(111)
    ax.set_aspect("equal")

    for screen_nr in G.nodes:
        sample_img_path = (
            f"src/appcommander/{package_name}/V{app_version}/verified/"
            + f"{screen_nr}.png"
        )
        if Path(sample_img_path).is_file():
            img = imread(sample_img_path)
        else:
            img = imread("Unknown_screen.resized")

        G.add_node(screen_nr, image=img)

    fig = plt.figure(figsize=(12, 3))
    ax = plt.subplot(111)

    horizontal_edges, vertical_edges = get_horizontal_and_vertical_edges(G=G)
    G.remove_edges_from(list(G.edges()))

    # draw vertical edges using a larger node size
    for edge in vertical_edges:
        if edge not in G.edges:
            G.add_edge(edge[0], edge[1])

    nx.draw_networkx_edges(G, pos, ax=ax, node_size=8000, arrowsize=30)

    # add the horizontal edges with a smaller node size
    for edge in horizontal_edges:
        if edge not in G.edges:
            G.add_edge(edge[0], edge[1])
    nx.draw_networkx_edges(G, pos, ax=ax, node_size=3000, arrowsize=30)

    set_node_images(G=G, package_name=package_name, app_version=app_version)

    trans = ax.transData.transform
    trans2 = fig.transFigure.inverted().transform

    piesize = 0.4  # this is the image size
    p2 = piesize / 2.0
    for n in G:
        xx, yy = trans(pos[n])  # figure coordinates
        xa, ya = trans2((xx, yy))  # axes coordinates
        a = plt.axes([xa - p2, ya - p2, piesize, piesize])
        a.set_aspect("equal")
        a.imshow(G.nodes[n]["image"])
        a.axis("off")
    ax.axis("off")
    plt.savefig(f"src/appcommander/{package_name}/V{app_version}/flow.png")


def set_cyclical_node_coords(
    evaluated_nodes: List[int],
    G: nx.DiGraph,
    start_nodename: int,
    y: int,
) -> nx.DiGraph:
    """Sets the coordinates for a cyclical graph."""
    neighbors = list(G.neighbors(start_nodename))
    if "pos" not in G.nodes[start_nodename].keys():
        raise KeyError("Error, pos value of start node was not set.")
    for x, neighbor in enumerate(neighbors):
        if neighbor not in evaluated_nodes:
            if "pos" not in G.nodes[neighbor].keys():
                G.nodes[neighbor]["pos"] = [y + 1, x]
    for x, neighbor in enumerate(neighbors):
        if neighbor not in evaluated_nodes:
            evaluated_nodes.append(neighbor)
            set_cyclical_node_coords(
                evaluated_nodes=evaluated_nodes,
                G=G,
                start_nodename=neighbor,
                y=y + 1,
            )


def set_node_images(
    G: nx.DiGraph, package_name: str, app_version: str
) -> None:
    """Sets the phone screens as node images to visualise the script flow
    between screens."""
    # Image URLs for graph nodes
    icons: Dict[str, str] = {}
    for screen_nr in G.nodes:
        img_path = (
            f"src/appcommander/{package_name}/V{app_version}/verified/"
            + f"{screen_nr}.png"
        )
        if Path(img_path).is_file():
            icons[screen_nr] = img_path
        else:

            icons[screen_nr] = "Unknown_screen.png"

        # TODO: verify file exists, use dummy file otherwise.
    # Load images
    images = {k: PIL.Image.open(fname) for k, fname in icons.items()}

    # Generate the computer network graph
    for nodename in G.nodes:
        G.nodes[nodename]["image"] = images[nodename]


def get_horizontal_and_vertical_edges(
    G: nx.DiGraph,
) -> Tuple[List[Tuple[int, int]], List[Tuple[int, int]]]:
    """Returns the vertical edges, and the horizontal edges."""
    horizontal: List[Tuple[int, int]] = []
    vertical: List[Tuple[int, int]] = []

    for edge in G.edges:
        left = edge[0]
        right = edge[1]
        if G.nodes[left]["pos"][1] != G.nodes[right]["pos"][1]:
            vertical.append(edge)
        else:
            horizontal.append(edge)
    return horizontal, vertical


def resize_image(
    input_filepath: str,
    target_height: int,
    target_width: int,
) -> None:
    """Stretch an image into a specific resolution."""

    if Path(input_filepath).is_file():
        outfile = os.path.splitext(input_filepath)[0] + ".resized"
        if input_filepath != outfile:
            try:
                img = Image.open(input_filepath)
                # im.thumbnail(size, Image.Resampling.LANCZOS)
                img = img.resize(
                    (target_width, target_height), Image.ANTIALIAS
                )
                img.save(outfile, "png")
            except OSError:
                print(f"Cannot resise image:{input_filepath}")
    else:
        raise FileNotFoundError(f"Error:{input_filepath}")


def get_existing_image_size(
    G: Union[nx.Graph, nx.DiGraph],
    package_name: str,
    app_version: str,
) -> Tuple[int, int]:
    """Loop through the existing screen png files, and returns the dimensions
    of the first screen png."""
    for screen_nr in G.nodes:
        img_path = (
            f"src/appcommander/{package_name}/V{app_version}/verified/"
            + f"{screen_nr}.png"
        )
        if Path(img_path).is_file():
            im = Image.open(img_path)
            width, height = im.size
            return width, height
    raise FileNotFoundError("Error, no screen found.")



"""After the "Connection Request" has been granted, the app welcomes the user
with screens 1,2,3,4."""
# pylint: disable=R0801
import inspect
from typing import TYPE_CHECKING, Callable, Dict, List, Union

import networkx as nx
from typeguard import typechecked
from uiautomator import AutomatorDevice

from appcommander.at_bitfire_davdroid.V4_2_6.helper import get_calendar_names
from appcommander.helper import get_screen_as_dict
from appcommander.org_torproject_android.V16_6_3_RC_1.helper import (
    get_torified_item_index_dict,
)
from appcommander.Screen import Screen
from appcommander.script_orientation import get_expected_screen_nrs

if TYPE_CHECKING:
    from appcommander.Script import Script
else:
    Script = object


@typechecked
def screen_8() -> Screen:
    """Done adding account, now sync and if not done yet, set permmisions."""

    max_retries = 3
    screen_nr = 8
    wait_time_sec = 1
    required_objects: List[Dict[str, str]] = [
        {
            "@text": "CARDDAV",
        },
        {
            "@text": "CALDAV",
        },
        {
            "@resource-id": "at.bitfire.davdroid:id/sync",
        },
        {
            "@resource-id": "at.bitfire.davdroid:id/list",
        },
    ]

    # pylint: disable=W0613
    @typechecked
    def get_next_actions(
        required_objects: List[Dict[str, str]],
        optional_objects: List[Dict[str, str]],
        script: Script,
    ) -> Union[Callable[[AutomatorDevice, Screen, Script], Dict], None]:
        """Looks at the required objects and optional objects and determines
        which actions to take next.
        An example of the next actions could be the following List:
        0. Select a textbox.
        1. Send some data to a textbox.
        2. Click on the/a "Next" button.

        Then the app goes to the next screen and waits a pre-determined
        amount, and optionally retries a pre-determined amount of attempts.
        """

        return actions_0

    return Screen(
        get_next_actions=get_next_actions,
        is_start=True,
        max_retries=max_retries,
        screen_nr=screen_nr,
        wait_time_sec=wait_time_sec,
        required_objects=required_objects,
    )


# pylint: disable=W0613
@typechecked
def actions_0(dev: AutomatorDevice, screen: Screen, script: Script) -> Dict:
    """Performs the actions in option 0 in this screen.

    For this screen, it waits until the phone is done querying the
    server.
    """

    required_object: Dict[str, str] = {
        "@resource-id": "at.bitfire.davdroid:id/list"
    }

    # Reload the screen data.
    unpacked_screen_dict: Dict = get_screen_as_dict(
        dev=dev,
        unpack=True,
        screen_dict={},
        reload=True,
    )

    # Get a subdict based on a value inside the dict.
    item_dict = get_torified_item_index_dict(
        required_object, unpacked_screen_dict, {}
    )

    cal_names = get_calendar_names(sub_screen_dict=item_dict)
    for calendar_name in cal_names:
        dev(text=calendar_name).click()

    # Press sync icon.
    dev(resourceId="at.bitfire.davdroid:id/sync").click()

    # Return the expected screens, using get_expected_screen_nrs.
    action_nr: int = int(inspect.stack()[0][3][8:])  # 8 for:len(actions__)
    screen_nr: int = screen.screen_nr
    script_flow: nx.DiGraph = script.script_graph
    return {
        "expected_screens": get_expected_screen_nrs(
            G=script_flow, screen_nr=screen_nr, action_nr=action_nr
        )
    }



"""Performs verifications on the status of the phone."""
from typing import TYPE_CHECKING, Dict, List, Tuple

import networkx as nx
from typeguard import typechecked
from uiautomator import AutomatorDevice

from appcommander.helper import (
    export_screen_data,
    get_screen_as_dict,
    is_expected_screen,
)
from appcommander.script_orientation import get_expected_screens

# pylint: disable=R0801
if TYPE_CHECKING:
    from appcommander.Screen import Screen
    from appcommander.Script import Script
else:
    Screen = object
    Script = object


@typechecked
def can_proceed(
    dev: AutomatorDevice,
    expected_screennames: List[int],
    retry: bool,
    script: Script,
) -> Tuple[bool, int]:
    """Checks whether the screen is expected, raises an error if not.

    And it returns the current screen number.
    """
    # get current screen dict.
    unpacked_screen_dict: Dict = get_screen_as_dict(
        dev=dev,
        unpack=True,
        screen_dict={},
        reload=False,
    )

    # verify current_screen in next_screens.
    is_expected, screen_nr = current_screen_is_expected(
        dev=dev,
        expected_screennames=expected_screennames,
        retry=retry,
        script_graph=script.script_graph,
        unpacked_screen_dict=unpacked_screen_dict,
    )

    # end_screens = get end_screens()
    if not is_expected:
        # Export the actual screen, screen data and expected screens in
        # specific error log folder.
        export_screen_data(
            dev=dev,
            screen_dict=unpacked_screen_dict,
            screen_nr=screen_nr,
            script=script,
            overwrite=True,
            subdir="error",
        )
        raise ReferenceError(
            f"Error, the expected screen was not found in:{screen_nr}. "
            + f"Searched for:{expected_screennames}. The accompanying screen "
            + "and xml can be found in:src/appcommander/<package_name>/<app_"
            + f"version>/error/{screen_nr}.json"
        )
    return is_expected, screen_nr


@typechecked
def current_screen_is_expected(
    dev: AutomatorDevice,
    expected_screennames: List[int],
    retry: bool,
    script_graph: nx.DiGraph,
    unpacked_screen_dict: Dict,
) -> Tuple[bool, int]:
    """Determines whether the current screen is one of the expected screens."""
    expected_screens: List[Screen] = get_expected_screens(
        expected_screennames, script_graph
    )
    for expected_screen in expected_screens:
        if is_expected_screen(
            dev=dev,
            expected_screen=expected_screen,
            retry=retry,
            unpacked_screen_dict=unpacked_screen_dict,
        ):

            return (
                True,
                int(str(expected_screen.screen_nr)),
            )
    return (False, -1)



"""Verifies the given CLI arguments are valid in combination with each
other."""

from typing import List

from typeguard import typechecked

from appcommander.helper import run_bash_command


@typechecked
def assert_phone_is_connected() -> None:
    """Throws error if phone is not connected via ADB."""
    # Launc the app on phone.
    command = "adb devices"
    output = run_bash_command(
        await_compilation=True, bash_command=command, verbose=False
    )
    lines = output.split("\n")

    found_device: bool = False
    # TODO: make more robust check, e.g. eat "List of devs attached" and see
    # whether any a-Z 0-9 characters exist in output.
    for line in lines:
        if "List of devices attached" not in line:
            for letter in line:
                if letter.isalnum():
                    found_device = True

    if not found_device:
        raise Exception("Error, no adb dev is found.")

    # TODO: check if more than one devs are connected, and if yes, raise
    # exception if user did not specify the desired dev name.


@typechecked
def assert_app_is_installed(package_name: str) -> None:
    """Throws error if the app is installed on the phone."""
    assert_phone_is_connected()
    command = "adb shell pm list packages"
    output = run_bash_command(
        await_compilation=True, bash_command=command, verbose=False
    )
    installed_package_list: List[str] = list(set(output.split("\n")))
    installed_packages = list(
        map(
            lambda x: x.replace("package:", "").replace("'", ""),
            installed_package_list,
        )
    )

    if package_name not in installed_packages:
        raise Exception(
            f"Error, the app:'{package_name}' with package name:'"
            + f"{package_name}', is not yet installed:"
            + f"{sorted(installed_packages)}"
        )


@typechecked
def assert_app_version_is_correct(package_name: str, app_version: str) -> None:
    """Throws error if the app version found on phone is not as expected."""
    assert_app_is_installed(package_name=package_name)

    print(f"TODO: assert app version is correct:{app_version}")



"""Functions to assist a script file for an arbitrary app."""
import importlib
from types import ModuleType
from typing import TYPE_CHECKING, Any, List, Optional, Tuple

import networkx as nx
from typeguard import typechecked

if TYPE_CHECKING:
    from appcommander.Screen import Screen
    from appcommander.Script import Script
else:
    Script = object
    Screen = object


@typechecked
def create_screens(script: Script) -> List[Screen]:
    """Creates the screens as networkx nodes."""

    modules, screen_func_names = load_screen_files_per_app_version(
        script.app_version_mod_path, script.script_graph
    )
    screens: List[Screen] = []

    # Create the Screen objects programmatically.
    for i, module in enumerate(modules):
        # Create the function (reference) programmatically.
        # module represents the file that contains the screen function.
        # screen_func_name[i] is the name of the screen_i function in that
        # module.
        # screen_function is the actual Pythonic function handle. it is
        # like writing: screen_3 if i==3.
        screen_function = getattr(module, screen_func_names[i])
        # execute the screen function, which returns a Screen object.
        screens.append(screen_function())

    # Add the screen objects to the script graph.
    for screen in screens:
        script.script_graph.nodes[screen.screen_nr]["Screen"] = screen
    return screens


@typechecked
def load_screen_files_per_app_version(
    app_version_mod_path: str, graph: nx.DiGraph
) -> Tuple[List[ModuleType], List[str]]:
    """A module is a python file (in this case).

    So this script loads the Python files from which the screen object
    data is loaded. Each app and version has its own screen_x files.
    """

    # Specify path to screen_files
    moduleNames = []
    screen_func_names = []
    modules = []
    for screen_index in range(0, len(graph)):
        module_name = f"{app_version_mod_path}screen_{screen_index}"
        moduleNames.append(module_name)
        screen_func_names.append(f"screen_{screen_index}")
        my_module = importlib.import_module(module_name)
        modules.append(my_module)
    return modules, screen_func_names


@typechecked
def load_script_attribute(  # type:ignore[misc]
    app_version_mod_path: str,
    filename: str,
    obj_name: str,
    attribute_name: Optional[str] = None,
) -> Any:
    """Loads an attribute for the script object.

    If attribute_name is empty, it will load the object named <filename>
    inside the <filename>, otherwise it will load the attribute
    <attribute_name> of that object.
    """

    file_path = f"{app_version_mod_path}{filename}"
    imported_file = importlib.import_module(file_path)

    the_object = getattr(imported_file, obj_name)
    if attribute_name is not None:
        attribute = getattr(the_object(), attribute_name)
        return attribute
    return the_object



"""Contains hardcoded data."""

# keys: The app name as the user sees it.
# values: The app name on the Android filesystem.
from typing import Dict

app_name_mappings: Dict[str, str] = {
    "Nextcloud": "org.nextcloud.android",
    "DAVx5": "at.bitfire.davdroid",
    "Orbot": "org.torproject.android",
}
# TODO: verify all supported package names are unique.



"""Completes the tasks specified in the arg_parser."""
import argparse
from typing import Dict

from typeguard import typechecked
from uiautomator import device

from appcommander.hardcoded import app_name_mappings
from appcommander.helper import export_screen_data, get_screen_as_dict
from appcommander.plot_script_flow import visualise_script_flow
from appcommander.run_script import run_script
from appcommander.Script import Script
from appcommander.verification.arg_verification import (
    get_verified_apps_to_torify,
    sort_out_app_name_and_package_name,
)
from appcommander.verification.verify_phone_connection import (
    assert_app_is_installed,
    assert_app_version_is_correct,
)


@typechecked
def process_args(args: argparse.Namespace) -> None:
    """Processes the arguments and ensures the accompanying tasks are
    executed."""
    app_name, package_name = sort_out_app_name_and_package_name(
        args.app_name, app_name_mappings=app_name_mappings
    )
    input_data: Dict = {}
    if args.torify:
        input_data["torifying_apps"] = get_verified_apps_to_torify(
            app_name_mappings, args.torify
        )
    if args.nextcloud_username:
        input_data["nextcloud_username"] = args.nextcloud_username
        input_data["nextcloud_password"] = args.nextcloud_password
        # TODO: prompt user for pwd. or safely communicate from
        # Collabora Online.
    if args.onion_url:
        input_data["onion_url"] = args.onion_url

    apk_script = Script(
        app_name=app_name,
        overwrite=False,
        package_name=package_name,
        cli_input_data=input_data,
        version=args.version,
    )
    if args.export_screen:
        unpacked_screen_dict: Dict = get_screen_as_dict(
            dev=device,
            unpack=True,
            screen_dict={},
            reload=False,
        )
        export_screen_data(
            dev=device,
            screen_dict=unpacked_screen_dict,
            screen_nr=args.export_screen,
            script=apk_script,
            overwrite=True,
            subdir="unverified",
        )
    elif args.export_script_flow:
        visualise_script_flow(
            G=apk_script.script_graph,
            package_name=apk_script.package_name.replace(".", "_"),
            app_version=apk_script.version.replace(".", "_").replace(" ", "_"),
        )
    else:
        print("")
        # pprint(apk_script.__dict__)
        # Also verifies phone is connected.
        assert_app_is_installed(package_name=package_name)
        assert_app_version_is_correct(
            package_name=package_name,
            app_version=args.version,
        )

        run_script(apk_script, device)



"""After the "Connection Request" has been granted, the app welcomes the user
with screens 1,2,3,4."""
# pylint: disable=R0801
import inspect
from typing import TYPE_CHECKING, Callable, Dict, List, Union

import networkx as nx
from typeguard import typechecked
from uiautomator import AutomatorDevice

from appcommander.Screen import Screen
from appcommander.script_orientation import get_expected_screen_nrs

if TYPE_CHECKING:
    from appcommander.Script import Script
else:
    Script = object


@typechecked
def screen_2() -> Screen:
    """Creates the settings for a starting screen where Orbot is not yet
    started."""

    max_retries = 1
    screen_nr = 2
    wait_time_sec = 0.1
    required_objects: List[Dict[str, str]] = [
        {
            "@text": "Browse the internet how you expect you should.",
        },
        {
            "@text": "No tracking. No censorship.",
        },
    ]

    # pylint: disable=W0613
    @typechecked
    def get_next_actions(
        required_objects: List[Dict[str, str]],
        optional_objects: List[Dict[str, str]],
        script: Script,
    ) -> Union[Callable[[AutomatorDevice, Screen, Script], Dict], None]:
        """Looks at the required objects and optional objects and determines
        which actions to take next.
        An example of the next actions could be the following List:
        0. Select a textbox.
        1. Send some data to a textbox.
        2. Click on the/a "Next" button.

        Then the app goes to the next screen and waits a pre-determined
        amount, and optionally retries a pre-determined amount of attempts.
        """

        return actions_0

    return Screen(
        get_next_actions=get_next_actions,
        is_start=True,
        max_retries=max_retries,
        screen_nr=screen_nr,
        wait_time_sec=wait_time_sec,
        required_objects=required_objects,
    )


# pylint: disable=W0613
@typechecked
def actions_0(dev: AutomatorDevice, screen: Screen, script: Script) -> Dict:
    """Performs the actions in option 0 in this screen.

    For this screen, it clicks the "Next" button (icon=">") in the
    bottom right.
    """
    dev(resourceId="org.torproject.android:id/next").click()

    # Return the expected screens, using get_expected_screen_nrs.
    action_nr: int = int(inspect.stack()[0][3][8:])
    screen_nr: int = screen.screen_nr
    script_flow: nx.DiGraph = script.script_graph
    return {
        "expected_screens": get_expected_screen_nrs(
            G=script_flow, screen_nr=screen_nr, action_nr=action_nr
        )
    }



"""Parses CLI arguments that specify on which platform to simulate the spiking
neural network (SNN)."""
import argparse

from typeguard import typechecked


@typechecked
def parse_cli_args() -> argparse.Namespace:
    """Reads command line arguments and converts them into python arguments."""
    # Instantiate the parser
    parser = argparse.ArgumentParser(
        description="Optional description for arg" + " parser"
    )

    # Create argument parsers to allow user to specify what to run.
    # Allow user run the experiment on a graph from file.
    parser.add_argument(
        "-a",
        "--app-name",
        action="store",
        type=str,
        help=(
            "The name of the Android app as used by your file system. E.g. "
            + "instead of 'Orbot' write: 'org.torproject.android'."
        ),
    )

    # Take a screenshot and store the UI information in a .json file.
    parser.add_argument(
        "-e",
        "--export-screen",
        action="store",
        type=int,
        help=(
            "Specify the screen number whose json data and .png data are "
            + "stored."
        ),
    )

    # Take a screenshot and store the UI information in a .json file.
    parser.add_argument(
        "-f",
        "--export_script_flow",
        action="store_true",
        help=("Show a graph  of the screen flow of this script for an app."),
    )

    # Create argument to allow user to specify which apps Orbot should torify.
    parser.add_argument(
        "-t",
        "--torify",
        action="store",
        type=str,
        help=(
            "The names (csv) of the Android apps that you want Orbot to torify"
            + ". You can choose from (left or right):'."
        ),
    )

    # Create argument to allow user to specify its nextcloud username.
    parser.add_argument(
        "-nu",
        "--nextcloud-username",
        action="store",
        type=str,
        help=("Your Nextcloud username."),
    )

    # Create argument to allow user to specify its nextcloud username.
    parser.add_argument(
        "-np",
        "--nextcloud-password",
        action="store",
        type=str,
        help=(
            "Your Nextcloud password. "
            + "TODO: build support for local safe passing."
        ),
    )

    # Create argument to allow user to specify your onion url.
    parser.add_argument(
        "-o",
        "--onion-url",
        action="store",
        type=str,
        help=("Your url like: lakjdsf2340usdffa.onion"),
    )

    # Run experiment on a particular experiment_settings json file.
    parser.add_argument(
        "-v",
        "--version",
        action="store",
        type=str,
        help=("Give the version of the Android app."),
    )

    # Load the arguments that are given.
    args = parser.parse_args()
    return args



"""Runs bash commands."""
import subprocess  # nosec
from typing import Union

from typeguard import typechecked


@typechecked
def run_bash_command(
    await_compilation: bool, bash_command: str, verbose: bool
) -> Union[None, str]:
    """Runs a bash command."""
    if await_compilation:
        if verbose:
            subprocess.call(bash_command, shell=True)  # nosec
        else:
            output = subprocess.check_output(  # nosec
                bash_command,
                shell=True,
                # stderr=subprocess.DEVNULL,
                # stdout=subprocess.DEVNULL,
            )
            return output.decode("utf-8")
    else:
        if verbose:
            # pylint: disable=R1732
            subprocess.Popen(bash_command, shell=True)  # nosec
        else:
            # pylint: disable=R1732
            subprocess.Popen(  # nosec
                bash_command,
                shell=True,
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
            )
    return None



"""Packaging logic for appcommander."""
from __future__ import annotations

import os
import sys

import setuptools

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "src"))
setuptools.setup(
    install_requires = [
        # Lists the pip packages used by the src code of this project. Doesn not
        # include pre-commit packages.
        # Parse xml files into Python dictionaries.
        "xmltodict",
        # Plot the script flow to png.
        "matplotlib",
        # Plot the script flow as a graph, to png.
        "networkx",
        # Allow for auto generation of type-hints during runtime.
        "pyannotate",
        # Run python tests.
        "pytest-cov",
        # Ensure the python function arguments are verified at runtime.
        "typeguard",
        # Control Android apps through their user interface (UI).
        "uiautomator",
        
        # Some dependencies for service-identity 21.1.0
        "pyasn1",
        "pyasn1-modules",
        "pyOpenSSL",
    ]
    # install_requires=['numpy','pandas'],
    # setup_requires=['pytest-runner'],
    # tests_require=['wandb','pytest==4.4.1'],
    # test_suite='tests',
)



"""Stores the flow logic of the script in a networkx graph."""

import networkx as nx
from typeguard import typechecked


class Screen_flow:
    """Contains a map from screen number and action number to next expected
    screens."""

    # pylint: disable=R0903
    @typechecked
    def __init__(
        self,
    ) -> None:
        """Creates the networkx graph with the screen nrs as nodes, and the
        action lists as edges."""
        self.G = nx.DiGraph()
        for nodename in list(range(0, 7 + 1)):
            self.G.add_node(nodename)
        self.G.add_edge(0, 1, actions=[0])

        self.G.add_edge(1, 2, actions=[0])
        self.G.add_edge(2, 3, actions=[0])
        self.G.add_edge(3, 4, actions=[0])
        self.G.add_edge(4, 5, actions=[0])
        self.G.add_edge(4, 7, actions=[0])

        self.G.add_edge(5, 6, actions=[0])
        # self.G.add_edge(5, 5, actions=[1])
        self.G.add_edge(5, 7, actions=[1])

        self.G.add_edge(6, 5, actions=[0])
        self.G.add_edge(6, 7, actions=[0])

        self.G.add_edge(7, 6, actions=[0])
